# coding: utf-8
##########################################################################
# NSAp - Copyright (C) CEA, 2019
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

# Third party import
import pandas as pd

# Package import
from .loader_base import LoaderBase


class TSV(LoaderBase):
    """ Define the TSV loader.
    """
    allowed_extensions = [".tsv"]

    def load(self, path):
        """ A method that load the table data.

        Parameters
        ----------
        path: str
            the path to the table to be loaded.

        Return
        ------
        data: pandas DataFrame
            the loaded table.
        """
        return pd.read_table(path, sep="\t")

    def save(self, data, outpath):
        """ A method that save the table.

        Parameters
        ----------
        data: pandas DataFrame
            the table to be saved.
        outpath: str
            the path where the the table will be saved.
        """
        data.to_csv(sep="\t")



# coding: utf-8
##########################################################################
# NSAp - Copyright (C) CEA, 2019
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

# System import
import os
import glob
import pickle
import datetime

# Third party import
import pandas as pd
from grabbit import Layout

# Package import
from .io import load

# Define global parameters
BASE_ENTITIES = ["subject", "session", "task", "run", "suffix"]


class Caravel(object):
    """ Object to retrieve data from a BIDS directory or a CubicWeb instance.
    """
    AVAILABLE_LAYOUTS = ("sourcedata", "derivatives", "phenotype")

    def __init__(self, project, layoutdir):
        """ Initialize the Caravel class.

        Parameters
        ----------
        project: str
            the name of the project you are working on.
        layoutdir: str
            the location of the pre-generated parsing representations. If None
            switch to managers mode.
        """
        self.project = project
        self.layouts = {}
        _conf = self._get_conf()
        if project not in _conf:
            raise ValueError(
                "Unkown configuration for project '{0}'. Available projects "
                "are: {1}.".format(project, _conf.keys()))
        self.conf = _conf[project]
        if layoutdir is not None:
            _repr = self._get_repr(layoutdir)
            if project not in _repr:
                raise ValueError(
                    "Unkown representation for project '{0}'. Available "
                    "projects are: {1}.".format(project, _repr.keys()))
            self.representation = _repr[project]
        else:
            self.representation = {}

    def _check_layout(self, name):
        """ Check if the layout name is supported.
        """
        if name not in self.AVAILABLE_LAYOUTS:
            raise ValueError(
                "Layout '{0}' is not yet supported. Available layouts are: "
                "{1}.".format(name, AVAILABLE_LAYOUTS))

    def _get_layout(self, name, layoutdir=None):
        """ Get a layout: chack if already loaded or try to load it from
        the latest representation. If not found,raise an error.
        """
        layout = self.layouts.get(name, None)
        if layout is None:
            if layoutdir is None:
                raise ValueError(
                    "Impossible to load layout '{0}', please specify a layout "
                    "directory.".format(name))

    def _get_conf(self):
        """ List all the configurations available and sort them by project.
        """
        confdir = os.path.join(
            os.path.abspath(os.path.dirname(__file__)), "conf")
        conf = {}
        for path in glob.glob(os.path.join(confdir, "*.conf")):
            basename = os.path.basename(path).replace(".conf", "")
            project, name = basename.split("_")
            if project not in conf:
                conf[project] = {}
            conf[project][name] = path
        return conf

    def _get_repr(self, layoutdir):
        """ List all the layout representation available and sort them by
        dates.
        """
        representations = {}
        for path in glob.glob(os.path.join(layoutdir, "*.pkl")):
            basename = os.path.basename(path).replace(".pkl", "")
            project, name, timestamp = basename.split("_")
            if project not in representations:
                representations[project] = {}
            representations[project].setdefault(name, []).append(
                {"date": timestamp, "path": path})
        for project, project_data in representations.items():
            for name, name_data in project_data.items():
                name_data.sort(key=lambda x: datetime.datetime.strptime(
                    x["date"], "%Y-%m-%d"))
        return representations        

    def _check_conf(self, name):
        """ Check if configuration is declared for the layout.
        """
        if name not in self.conf:
            raise ValueError(
                "No configuration available for layout '{0}'. Please contact "
                "the module developpers to add the support for your project.")

    def _load_layout(self, name):
        """ Load a loayout from its pre-generated represnetation.
        """
        if name not in self.layouts:
            if name not in self.representation:
                raise ValueError(
                    "A pre-generated '{0}' layout for your project '{1}' is "
                    "expected in user mode. Please contact the developpers "
                    "of the module.".format(name, self.project))
            path = self.representation[name][-1]["path"]
            with open(path, "rb") as open_file:
                self.layouts[name] = pickle.load(open_file)
        return self.layouts[name]

    def export_layout(self, name):
        """ Export a layout as a pandas DataFrame.

        Parameters
        ----------
        name: str
            the name of the layout.

        Returns
        -------
        df: pandas DataFrame
            the converted layout.
        """
        layout = self._load_layout(name)
        return layout.as_data_frame()

    def list_keys(self, name):
        """ List all the filtering keys available in the layout.

        Parameters
        ----------
        name: str
            the name of the layout.

        Returns
        -------
        keys: list
            the layout keys.
        """
        layout = self._load_layout(name)
        return list(layout.entities.keys())

    def list_values(self, name, key):
        """ List all the filtering key values available in the layout.

        Parameters
        ----------
        name: str
            the name of the layout.
        key: str
            the name of key in the layout.

        Returns
        -------
        values: list
            the key assocaited values in the layout.
        """
        layout = self._load_layout(name)
        if key not in layout.entities:
            raise ValueError("Unrecognize layout key '{0}'.".format(key))
        return list(layout.unique(key))

    def filter_layout(self, name, extensions=None, **kwargs):
        """ Filter the layout by using a combination of key-values rules.

        Parameters
        ----------
        name: str
            the name of the layout.
        extensions: str or list of str
            a filtering rule on the file extension.
        kwargs: dict
            the filtering options.

        Returns
        -------
        df: pandas DataFrame
            the filtered layout.
        """
        layout = self._load_layout(name)
        if extensions is not None:
            kwargs["extensions"] = extensions
        header = None
        files = layout.get(**kwargs)
        if len(files) == 0:
            df = pd.DataFrame()
        else:
            file_obj = files[0]
            header = ["filename"]
            for key in layout.entities:
                if hasattr(file_obj, key):
                    header.append(key)
            data = []
            for file_obj in files:
                row = []
                for key in header:
                    row.append(getattr(file_obj, key))
                data.append(row)
            df = pd.DataFrame(data, columns=header)
        return df           

    def load_data(self, name, df):
        """ Load the data contained in the filename column of a pandas
        DataFrame.

        Note:
        Only a couple of file extensions are supported. If no loader has been
        found None is returned.

        Parameters
        ----------
        name: str
            the name of the layout.
        df: pandas DataFrame
            a table with one 'filename' column.

        Returns
        -------
        data: dict
            a dictionaray containing the loaded data.
        """
        if "filename" not in df:
            raise ValueError("One 'filename' column expected in your table.")
        layout = self._load_layout(name)
        data = {}
        for path in df["filename"]:
            try:
                _data = load(path)
            except:
                _data = None
            if isinstance(_data, pd.DataFrame):
                file_obj = layout.files[path]
                for ent_name, ent_val in file_obj.entities.items():
                    if ent_name in BASE_ENTITIES:
                        _data[ent_name] = ent_val
                _data["dtype"] = name
                if "participant_id" in _data:
                    _data["participant_id"] = _data["participant_id"].str.replace(
                        "sub-", "")
            data[path] = _data
        return data

    def pickling_layout(self, bids_root, name, outdir):
        """ Load the requested BIDS layout and save it as a pickle.

        Parameters
        ----------
        bids_root: str
            path to the BIDS folder.
        name: str
            the name of subfolder to be parsed (the layout name).
        outdir: str
            the folder where the pickle will be generated.

        Returns
        -------
        outfile: str
            the generated layout representation location.
        """
        self._check_layout(name)
        self._check_conf(name)
        layout_root = os.path.join(bids_root, name)
        if not os.path.isdir(layout_root):
            raise ValueError("'{0}' is not a valid directory.")
        layout = Layout(layout_root, self.conf[name])
        self.layouts[name] = layout
        now = datetime.datetime.now()
        timestamp = "{0}-{1}-{2}".format(now.year, now.month, now.day)
        outfile = os.path.join(
            outdir, "{0}_{1}_{2}.pkl".format(self.project, name, timestamp))
        with open(outfile, "wb") as open_file:
            pickle.dump(layout, open_file, -1)
        return outfile




# coding: utf-8
##########################################################################
# NSAp - Copyright (C) CEA, 2019
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
This module defines common loaders imaging-genetic datasets.
"""

from .nifti import NIFTI
from .tsv import TSV



# coding: utf-8
##########################################################################
# NSAp - Copyright (C) CEA, 2019
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

# Package import
from caravel.loaders import TSV
from caravel.loaders import NIFTI

# Global parameters
# > define all the available loaders
LOADERS = [TSV, NIFTI]


def load(path, **kwargs):
    """ Load the data.

    Parameters
    ----------
    path: str
        the path to the data to be loaded.

    Returns
    -------
    data: object
        the loaded data.
    """
    loader = get_loader(path)
    return loader.load(path, **kwargs)


def save(data, path, **kwargs):
    """ Save the data.

    Parameters
    ----------
    data: object
        the data to be saved.
    path: str
        the destination file.
    """
    saver = get_saver(path)
    saver.save(data, path, **kwargs)


def get_loader(path):
    """ Search for a suitable loader in the declared loaders.
    Raise an exception if no loader is found.

    Parameters
    ----------
    path: str
        the path to the data to be loaded.

    Returns
    -------
    loader: @instance
        the loader instance.
    """
    for loader_class in LOADERS:
        loader = loader_class()
        if loader.can_load(path):
            return loader
    raise Exception("No loader available for '{0}'.".format(path))


def get_saver(path):
    """ Search for a suitable saver in the declared savers.
    Raise an exception if no saver is found.

    Parameters
    ----------
    path: str
        the path to the data to be saved.

    Returns
    -------
    saver: @instance
        the loader instance.
    """
    for saver_class in LOADERS:
        saver = saver_class()
        if saver.can_save(path):
            return saver
    raise Exception("No saver available for '{0}'.".format(path))



# coding: utf-8
##########################################################################
# NSAp - Copyright (C) CEA, 2019
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

"""
This module checks that all the dependencies are installed properly.
"""

# System import
import importlib
import distutils

# Package import
from .info import __version__
from .info import REQUIRES
from .info import LICENSE
from .info import AUTHOR


def _check_python_versions():
    """ Check that all the Python dependencies are satisfied.

    A dependency is expected to be formatted as follows:
    <mod_name>>=<mod_min_version>

    Returns
    -------
    versions: dict with 2-uplet
        the minimum required version and the installed version for each module.
        '?' means no package found.
    """
    versions = {}
    for dependency in REQUIRES:
        if ">=" in dependency:
            operator = ">="
        elif "==" in dependency:
            operator = "=="
        else:
            raise ValueError("'{0}' dependency no formatted correctly.".format(
                dependency))
        mod_name, mod_min_version = dependency.split(operator)
        if mod_name == "progressbar2":
            mod_name = "progressbar"
        try:
            mod_install_version = importlib.import_module(mod_name).__version__
        except:
            mod_install_version = "?"
        versions[mod_name] = (operator + mod_min_version, mod_install_version)
    return versions


def logo():
    """ pySAP logo is ascii art using fender font.

    Returns
    -------
    logo: str
        the pysap logo.
    """
    logo = r"""
   _ __    _  _                                                     _
  | '_ \  | || |   __     __ _      _ _   __ _    __ __    ___     | |
  | .__/   \_, |  / _|   / _` |    | '_| / _` |   \ V /   / -_)    | |
  |_|__   _|__/   \__|_  \__,_|   _|_|_  \__,_|   _\_/_   \___|   _|_|
_|'''''|_| ''''|_|'''''|_|'''''|_|'''''|_|'''''|_|'''''|_|'''''|_|'''''|
'`-0-0-''`-0-0-''`-0-0-''`-0-0-''`-0-0-''`-0-0-''`-0-0-''`-0-0-''`-0-0-'
    """
    return logo


def info():
    """ Dispaly some usefull information about the package.

    Returns
    -------
    info: str
        package information.
    """
    dependencies = "Dependencies: \n\n"
    dependencies_info = _check_python_versions()
    for name, (min_version, install_version) in dependencies_info.items():
        dependencies += "{0:15s}: {1:9s} - required | {2:9s} installed".format(
            name, min_version, install_version)
        dependencies += "\n"
    version = "Package version: {0}\n\n".format(__version__)
    license = "License: {0}\n\n".format(LICENSE)
    authors = "Authors: \n{0}\n".format(AUTHOR)
    return logo() + "\n\n" + version + license + authors + dependencies



# coding: utf-8
##########################################################################
# NSAp - Copyright (C) CEA, 2019
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################

# Module current version
version_major = 0
version_minor = 0
version_micro = 1

# Expected by setup.py: string of form "X.Y.Z"
__version__ = "{0}.{1}.{2}".format(version_major, version_minor, version_micro)

# Expected by setup.py: the status of the project
CLASSIFIERS = ["Development Status :: 1 - Planning",
               "Environment :: Console",
               "Operating System :: OS Independent",
               "Programming Language :: Python",
               "Topic :: Scientific/Engineering"]

# Project descriptions
description = """
pycaravel:  Python package that enables you to parse various source of data.
"""
SUMMARY = """
.. container:: summary-carousel

    pycaravel is a Python module for **data searching** that offers:

    * a common API for parsing multiple source of data (BIDS, CubicWeb, ...).
    * a common API to search in those datasets.
    * some utilities to load the retrived data.
"""
long_description = """
=========
pycaravel
=========

This module has been created to simplify the search of datasets in a BIDS
directory or a CubicWeb instance.
"""

# Main setup parameters
NAME = "pycaravel"
ORGANISATION = "CEA"
MAINTAINER = "Antoine Grigis"
MAINTAINER_EMAIL = "antoine.grigis@cea.fr"
DESCRIPTION = description
LONG_DESCRIPTION = long_description
EXTRANAME = "pycaravel webPage"
EXTRAURL = "http://cosmic.cosmostat.org/"
URL = "https://github.com/neurospin/pycaravel"
DOWNLOAD_URL = "https://github.com/neurospin/pycaravel"
LICENSE = "CeCILL-B"
CLASSIFIERS = CLASSIFIERS
AUTHOR = """
Antoine Grigis <antoine.grigis@cea.fr>
"""
AUTHOR_EMAIL = "antoine.grigis@cea.fr"
PLATFORMS = "Linux,OSX"
ISRELEASE = True
VERSION = __version__
PROVIDES = ["caravel"]
REQUIRES = [
    "pandas>=0.19.2",
    "grabbit>=0.2.5"
]
EXTRA_REQUIRES = {}



# coding: utf-8
##########################################################################
# NSAp - Copyright (C) CEA, 2019
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################


class LoaderBase(object):
    """ Base class for all loaders.
    """
    allowed_extensions = []

    def can_load(self, path):
        """ A method checking the file extension.

        Parameters
        ----------
        path: str
            the path to the data to be loaded.

        Return
        ------
        out: bool
            True if the file extension is valid, False otherwise.
        """
        for ext in self.allowed_extensions:
            if path.endswith(ext):
                return True
        return False

    def load(self, path):
        """ A method that load the data and associated metadata.

        Parameters
        ----------
        path: str
            the path to the data to be loaded.

        Return
        ------
        image: Data
            the loaded data.
        """
        raise NotImplementedError(
            "The 'load' method must be implemented in subclasses.")

    def can_save(self, outpath):
        """ A method checking the output file extension.

        Parameters
        ----------
        outpath: str
            the path where the the image will be saved.

        Return
        ------
        out: bool
            True if the output file extension is valid, False otherwise.
        """
        for ext in self.allowed_extensions:
            if outpath.endswith(ext):
                return True
        return False

    def save(self, data, outpath):
        """ A method that save the data and associated metadata.

        Parameters
        ----------
        image: Data
            the data to be saved.
        outpath: str
            the path where the the data will be saved.
        """
        raise NotImplementedError(
            "The 'save' method must be implemented in subclasses.")



#! /usr/bin/env python
##########################################################################
# NSAp - Copyright (C) CEA, 2019
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
##########################################################################


# System import
from __future__ import print_function
import os
import re
import sys
import platform
import subprocess
from pprint import pprint
from distutils.version import LooseVersion
from setuptools.command.build_ext import build_ext
from setuptools import setup, find_packages, Extension
from setuptools.command.test import test as TestCommand


# Package information
release_info = {}
infopath = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "caravel", "info.py"))
with open(infopath) as open_file:
    exec(open_file.read(), release_info)
pkgdata = {
    "caravel": [
        os.path.join("conf", "*.conf")]
}

# Write setup
setup(
    name=release_info["NAME"],
    description=release_info["DESCRIPTION"],
    long_description=release_info["LONG_DESCRIPTION"],
    license=release_info["LICENSE"],
    classifiers=release_info["CLASSIFIERS"],
    author=release_info["AUTHOR"],
    author_email=release_info["AUTHOR_EMAIL"],
    version=release_info["VERSION"],
    url=release_info["URL"],
    packages=find_packages(exclude="doc"),
    platforms=release_info["PLATFORMS"],
    extras_require=release_info["EXTRA_REQUIRES"],
    install_requires=release_info["REQUIRES"],
    package_data=pkgdata
)
